
//========================================================================================
//
//		Fun with prime numbers - the sieve of Eratosthenes v2.2.2
//		by Dan Brown
//		www.tacticaltinkering.co.uk
//
//========================================================================================


#include <iostream>
#include <Windows.h>

#define MAX		15485863				//15485863  //179424673 //982451653 //2038074743

#if MAX > 4294967295					// If we can, use 32 bit integers (twice as fast on a 32 bit system)
typedef unsigned __int64 BigInt;
#else
typedef unsigned int BigInt;
#endif

using namespace std;

//-----------------------------------------------------------------------------------------
// Name: CTimer [CLASS]
// Desc: A class for use as a fairly accurate stopwatch
//-----------------------------------------------------------------------------------------
class CTimer	{
	private:
		__int64 start;
		__int64 end;
		__int64 iElapsed()
		{
			LARGE_INTEGER lInt;
			QueryPerformanceCounter(&lInt);
			return (lInt.QuadPart-start);
		}
	public:
		double value;
		double valueMS;
		bool Start()
		{
			LARGE_INTEGER lInt;
			if(!QueryPerformanceCounter(&lInt))
			{
				std::cout << "\nTimer failed to start!";
				return false;
			}
			start = lInt.QuadPart;
			return true;
		}
		bool Stop()
		{
			LARGE_INTEGER lIntPF, lInt;
			if(!QueryPerformanceFrequency(&lIntPF)) return false;
			if(!QueryPerformanceCounter(&lInt)) return false;			
			end = lInt.QuadPart-start;
			value = (double)iElapsed()/(double)lIntPF.QuadPart;
			valueMS = value*1000;
			return true;
		}
		double TimeElapsed()
		{
			Stop();
			return value;
		}
} Timer;
//-----------------------------------------------------------------------------------------

//-----------------------------------------------------------------------------------------
// The number of [6k+1, 6k+5] pairs we need to cover all candidates up to MAX is 2*((MAX-MAX%6)/6+1)-1
// - Ignores 1, so the first element corresponds to 5
// - If MAX is of the form 6k+1 (or +2, +3, or +4), then the last element is superfluous, so we -1
// - If MAX is of the form 6k, then the last 2 elements are superfluous, so we -1 again
//-----------------------------------------------------------------------------------------
const BigInt arrSize = 2*((MAX-MAX%6)/6+1) - 1 - (MAX%6 < 5 ? 1 : 0) - (MAX%6 == 0 ? 1 : 0);	

char *n = new char[arrSize];					// Create the main array on the heap

int main()
{
	Timer.Start();
	
	std::fill(n, n + arrSize, 1);				// Assume all candidates are prime to start, so set all array values to 1

	cout << "\nInitialised the array (" << Timer.TimeElapsed() << "s)\n\nGenerating primes with the sieve of Eratosthenes ...\n\n";

	BigInt p = 5;			// p = current prime we're deleting multiples of in the bit array - ignoring 2!
	BigInt m = p*p;			// m = multiples of p greater than (p-1)*p, smaller multiples are already removed by smaller primes
	int r = p%6;			// r = remainders mod 6, to differentiate between 6k+1, 6k+3, 6k+5 forms

	while(m <= MAX)
	{
		//-----------------------------------------------------------------------------------------------------------------------
		// Loop through the array and remove all multiples of p
		//-----------------------------------------------------------------------------------------------------------------------
		if( r == 1 )										// r = 1: m%6 follow a repeating 1, 3, 5 pattern
		{
			while(m <= MAX) 
			{
				if(r == 1)			n[2*(m-1)/6-1] = 0;		// Of the form 6k+1
				else if(r == 5)		n[2*(m-5)/6] = 0;		// Of the form 6k+5

				m += 2*p;									// Ignore even multiples
				r += 2;
				if(r > 6) r -= 6;							// After r = 5 we want r = 1
			}
		}
		else												// r = 5: m%6 follow a repeating 5, 3, 1 pattern
		{
			r = 1;											// Starting multiple is m = p^2, and p^2 = 1 (mod 6) for any prime p
			while(m <= MAX) 
			{
				if(r == 1)			n[2*(m-1)/6-1] = 0;		// Of the form 6k+1
				else if(r == 5)		n[2*(m-5)/6] = 0;		// Of the form 6k+5

				m += 2*p;									// Ignore even multiples
				r -= 2;
				if(r < 0) r += 6;							// After r = 1 we want r = 5
			}
		}

		//---------------------------------------------------------------------
		// Find the next number left in the array. To do this, we:
		// - convert p into its position in the array
		// - increment p until we find a non-zero entry
		// - convert p back into the number the position represents
		//---------------------------------------------------------------------
		r = p%6;
		if(r == 1)			p = 2*(p-1)/6-1;
		else if(r == 5)		p = 2*(p-5)/6;
		else				p = -1;				// Only case left is when p = 3

		do {p++;} while( n[p] != 1 );

		p = p%2 == 0 ?	p*6/2+5 : (p-1)*6/2+7;
		//---------------------------------------------------------------------

		m = p*p;								// Starting multiple of p
		r = p%6;								// To find out if p is of the form 6k+1 or 6k+5
	}

	Timer.Stop();
	
	BigInt primesFound = 2;						// Not forgetting 2 and 3!

	for( BigInt a = 0 ; a < arrSize ; a++ )
	{
		if(n[a]==1)
		{
			primesFound++;

			if(primesFound==1000000) cout << "\n\nMillionth prime:\t" << (a%2 == 0 ? a*6/2+5 : (a-1)*6/2+7);
			if(primesFound==10000000) cout << "\n10 millionth prime:\t" << (a%2 == 0 ? a*6/2+5 : (a-1)*6/2+7);
			if(primesFound==50000000) cout << "\n50 millionth prime:\t" << (a%2 == 0 ? a*6/2+5 : (a-1)*6/2+7);
			if(primesFound==100000000) cout << "\n100 millionth prime:\t" << (a%2 == 0 ? a*6/2+5 : (a-1)*6/2+7);
		}
	}

	cout << "\n\nPrimes found:\t\t" << primesFound << "\nSearched up to:\t\t" << MAX << "\n\nTime elapsed:\t\t";
	Timer.value > 1 ? cout << Timer.value << "s" : cout << Timer.valueMS << "ms";
	
	cin.ignore();

	delete [] n;
}
